/* ------------------------------------------------ */
/* version of config.h for OS/2                     */
/* ------------------------------------------------ */

/* Define to empty if the keyword does not work.  */
#undef const

/* Define to `unsigned' if <sys/types.h> doesn't define.  */
#undef size_t

/* Define if you have the ANSI C header files.  */
#define STDC_HEADERS 1

/* Define if you have the <malloc.h> header file.  */
#define HAVE_MALLOC_H 1

/* Define if you have the <string.h> header file.  */
#define HAVE_STRING_H 1

/* Define if you have the <sys/types.h> header file.  */
#define HAVE_SYS_TYPES_H 1

/* Define if you have <alloca.h> and it should be used (not on Ultrix).  */
#define HAVE_ALLOCA_H

/* Define if platform-specific command line handling is necessary.  */
#define NEED_ARGV_FIXUP
#define argv_fixup(ac,av) { _response(ac,av); _wildcard(ac,av);}
